clear all ;
tc.ACCEL_S0 = 0 ;
tc.ACCEL_S1 = 1 ;
tc.ACCEL_S2 = 2 ;
tc.ACCEL_S3 = 3 ;
tc.ACCEL_S4 = 4 ;
tc.ACCEL_S5 = 5 ;
tc.ACCEL_S6 = 6 ;
tc.ACCEL_S7 = 7 ;

tc.accel_state      = tc.ACCEL_S0 ;
tc.cycle_time       = 0.006;

tc.jerk     = 2000 ;
tc.maxaccel = 200 ;
tc.maxvel   = 1000 ;

tc.reqvel   =100 ;
tc.target   = 100 ;

tc.feed_override = 1;
tc.cur_accel= 0;
tc.cur_vel=0;
tc.progress = 0;

tc.jerk = tc.jerk*tc.cycle_time^3 ;
tc.maxaccel = tc.maxaccel*tc.cycle_time^2 ;
tc.maxvel = tc.maxvel*tc.cycle_time^1 ;


tc.tc_target =  tc.target ;
tc.ts=0;
tc.ti=0;
tc.k=0; 
tc.s6_a=0 ;
tc.s6_v=0 ;
tc.s6_p=0 ;
tc.error_d=0 ;
tc.prev_s=0 ;
tc.prev_v=0;
tc.c1=0 ;
tc.c2=0 ;
tc.c3=0 ;
tc.c4=0 ;
tc.c5=0 ;
tc.c6=0;
tc.pi = vpa(pi,15);

tc.isend=0 ;

% 绘图 
f = figure("Name","S型曲线") ;
ax_a = subplot(4, 1, 1) ;
ax_v = subplot(4, 1, 2) ;
ax_s = subplot(4, 1, 3) ;
ax_g = subplot(4, 1, 4) ;

a_data =[0];
v_data =[0];
s_data =[0];
t_data = [0];
g_data = [0];
ts =0 ;
plot_handle_a = plot(ax_a,t_data,a_data, 'o');
plot_handle_v = plot(ax_v,t_data,v_data, 'o');
plot_handle_s = plot(ax_s,t_data,s_data, 'o');
plot_handle_g = plot(ax_g,t_data,g_data);

isend = tc.isend ;

while( ~isend)
    ts = tc.cycle_time + ts ;
    tc = ScurveUpdate(tc);
    isend = tc.isend ;
    % 更新数据
    curr_a = tc.cur_accel ;
    curr_v = tc.cur_vel ;
    curr_s = tc.progress ;
    curr_g = tc.accel_state ;

    t_data = [t_data, ts];

    a_data = [a_data, curr_a];
    v_data = [v_data, curr_v];
    s_data = [s_data, curr_s];
    g_data = [g_data, curr_g];

    % 更新 plot 对象的数据
    set(plot_handle_a, 'XData', t_data, 'YData', a_data);
    set(plot_handle_v, 'XData', t_data, 'YData', v_data);
    set(plot_handle_s, 'XData', t_data, 'YData', s_data);
    set(plot_handle_g, 'XData', t_data, 'YData', g_data);
    % 强制刷新图形窗口
    drawnow;
    
    % 添加短暂的延迟以观察动态效果
     pause(0.1);
end




% 函数
function tc = ScurveUpdate(tp)
    tc = tp;
    flag =1 ;
   %由于matlab 无法在其他地方使用 break
    while(flag)
    switch (tc.accel_state) 
    case tc.ACCEL_S0
        %  AT = AT + JT
        %  VT = VT + AT + 1/2JT
        %  PT = PT + VT + 1/2AT + 1/6JT
        tc.cur_accel = tc.cur_accel + tc.jerk;
        tc.cur_vel = tc.cur_vel + tc.cur_accel + 0.5 * tc.jerk;
        tc.progress = tc.progress + tc.cur_vel + 0.5 * tc.cur_accel + 1.0/6.0 * tc.jerk;
        
        % check if we hit accel limit at next BP
        if ((tc.cur_accel + tc.jerk) >= tc.maxaccel) 
            tc.cur_accel = tc.maxaccel;
            tc.accel_state = tc.ACCEL_S1;
           break ;
        end
        % check if we will hit velocity limit; 
        % assume we start decel from here to "accel == 0"
        %
        % AT = A0 + JT (let AT = 0 to calculate T)
        % VT = V0 + A0T + 1/2JT2
        t = ceil(tc.cur_accel / tc.jerk);
        req_vel = tc.reqvel * tc.feed_override * tc.cycle_time;
        if (req_vel > tc.maxvel) 
            req_vel = tc.maxvel;
        end
        vel = req_vel - tc.cur_accel * t + 0.5 * tc.jerk * t * t;
        if (tc.cur_vel >= vel) 
            tc.accel_state = tc.ACCEL_S2;
            break;
        end
        % check if we will hit progress limit
        % AT = AT + JT
        % VT = V0 + A0T + 1/2 JT^2
        % PT = P0 + V0T + 1/2A0T^2 + 1/6JT^3
        % distance for S2
        t = ceil(tc.cur_accel/tc.jerk);
        dist = tc.progress + tc.cur_vel * t + 0.5 * tc.cur_accel * t * t - 1.0/6.0 * tc.jerk * t * t * t;
        vel = tc.cur_vel + tc.cur_accel * t - 0.5 * tc.jerk * t * t;
        % distance for S3
        dist = dist+ (vel);

          
        %  0.5 * vel = vel + 0 * t1 - 0.5 * j * t1 * t1;
        % t1 = sqrt(vel/j)

        t = ceil(sqrt(vel/tc.jerk));
        % AT = AT + JT
        t1 = ceil(tc.maxaccel / tc.jerk);   % max time for S4
        if (t > t1) 
            % S4 . S5 . S6
            dist =dist+ t1 * vel;    % dist of (S4 + S6)
            % calc decel.dist for ACCEL_S5
           % t: time for S5
            t = (vel - tc.jerk * t1 * t1) / tc.maxaccel;
            v1 = vel - 0.5 * tc.jerk * t1 * t1;
            % PT = P0 + V0T + 1/2A0T^2 + 1/6JT^3
            dist = dist+ (v1 * t - 0.5 * tc.maxaccel * t * t);
        else
            % S4 . S6
            dist =dist+ t * vel;    % dist of (S4 + S6)
        end

        if (tc.tc_target < dist) 
            tc.accel_state = tc.ACCEL_S2;
            break;
        end

    case tc.ACCEL_S1
        % jerk is 0 at this state
        % VT = VT + AT + 1/2JT
        % PT = PT + VT + 1/2AT + 1/6JT
        tc.cur_vel = tc.cur_vel + tc.cur_accel;
        tc.progress = tc.progress + tc.cur_vel + 0.5 * tc.cur_accel;
        
        % check if we will hit velocity limit; 
        % assume we start decel from here to "accel == 0"
        %
        % AT = A0 + JT (let AT = 0 to calculate T)
        % VT = V0 + A0T + 1/2JT2
        % t = ceil(tc.cur_accel / tc.jerk);
        t = tc.cur_accel / tc.jerk;
        req_vel = tc.reqvel * tc.feed_override * tc.cycle_time;
        if (req_vel > tc.maxvel) 
            req_vel = tc.maxvel;
        end
        vel = req_vel - tc.cur_accel * t + 0.5 * tc.jerk * t * t;
        if (tc.cur_vel >= vel) 
            tc.accel_state = tc.ACCEL_S2;
            break;
        end
        
        % check if we will hit progress limit
        % distance for S2
        t = ceil(tc.cur_accel/tc.jerk);
        dist = tc.progress + tc.cur_vel * t + 0.5 * tc.cur_accel * t * t- 1.0/6.0 * tc.jerk * t * t * t;
        vel = tc.cur_vel + tc.cur_accel * t - 0.5 * tc.jerk * t * t;
        % distance for S3
        dist =dist+(vel);
        
        
        %0.5 * vel = vel + 0 * t1 - 0.5 * j * t1 * t1;
        %t1 = sqrt(vel/j)
        
        t = ceil(sqrt(vel/tc.jerk));
        % AT = AT + JT
        t1 = ceil(tc.maxaccel / tc.jerk);   % max time for S4
        if (t > t1) 
            % S4 . S5 . S6
            dist =dist+ t1 * vel;    % dist of (S4 + S6)
            % calc decel.dist for ACCEL_S5
            % t: time for S5
            t = (vel - tc.jerk * t1 * t1) / tc.maxaccel;
            v1 = vel - 0.5 * tc.jerk * t1 * t1;
            % PT = P0 + V0T + 1/2A0T^2 + 1/6JT^3
            dist =dist+ (v1 * t - 0.5 * tc.maxaccel * t * t);
         else 
            % S4 . S6
            dist =dist+ t * vel;    % dist of (S4 + S6)
        end
        
        if (tc.tc_target < dist) 
            tc.accel_state = tc.ACCEL_S2;
            break;
        end
    case tc.ACCEL_S2
        % to DECELERATE to ACCEL==0
        
        % AT = AT + JT
        % VT = VT + AT + 1/2JT
        % PT = PT + VT + 1/2AT + 1/6JT
        tc.cur_accel = tc.cur_accel - tc.jerk;
        tc.cur_vel = tc.cur_vel + tc.cur_accel - 0.5 * tc.jerk;
        tc.progress = tc.progress + tc.cur_vel + 0.5 * tc.cur_accel - 1.0/6.0 * tc.jerk;
        
        % check if (accel <= 0) at next BP
        acc = tc.cur_accel - tc.jerk;
        if (acc <= 0) 
            tc.accel_state = tc.ACCEL_S3;
            break;
        end
        
        % check if we will hit velocity limit at next BP
        req_vel = tc.reqvel * tc.feed_override * tc.cycle_time;
        if (req_vel > tc.maxvel) 
            req_vel = tc.maxvel;
        end
        % vel: velocity at next BP 
        vel = tc.cur_vel + tc.cur_accel - 1.5 * tc.jerk;
        if (vel > req_vel) 
            tc.cur_vel = req_vel;
            tc.accel_state = tc.ACCEL_S3;
            break;
        end
        
        % check if we will hit progress limit
        % refer to 2011-10-17 ysli design note
        % AT = AT + JT
        % VT = V0 + A0T + 1/2 JT^2
        % PT = P0 + V0T + 1/2A0T^2 + 1/6JT^3
        vel = tc.cur_vel;
        % distance for S3
        dist = tc.progress + (vel);
        
        %  0.5 * vel = vel + 0 * t1 - 0.5 * j * t1 * t1;
        %  t1 = sqrt(vel/j)
        t = ceil(sqrt(vel/tc.jerk));
        % AT = AT + JT
        t1 = ceil(tc.maxaccel / tc.jerk);   % max time for S4
        if (t > t1) 
            % S4 . S5 . S6
            dist =dist+ t1 * vel;    % dist of (S4 + S6)
            % calc decel.dist for ACCEL_S5
            % t: time for S5
            t = (vel - tc.jerk * t1 * t1) / tc.maxaccel;
            v1 = vel - 0.5 * tc.jerk * t1 * t1;
            % PT = P0 + V0T + 1/2A0T^2 + 1/6JT^3
            dist =dist+ (v1 * t - 0.5 * tc.maxaccel * t * t);
        else 
            % S4 . S6
            dist =dist+ t * vel;    % dist of (S4 + S6)
        end
        
        if (tc.tc_target < dist) 
            tc.accel_state = tc.ACCEL_S3;
            break;
        end
    case tc.ACCEL_S3
        % PT = PT + VT + 1/2AT + 1/6JT
        % , where (jerk == 0) and (accel == 0)
        tc.cur_accel = 0;
        tc.progress = tc.progress + tc.cur_vel;
        
        % check if we will hit progress limit
        % refer to 2011-10-17 ysli design note
        % AT = AT + JT
        % VT = V0 + A0T + 1/2 JT^2
        % PT = P0 + V0T + 1/2A0T^2 + 1/6JT^3

        %0.5 * vel = vel + 0 * t1 - 0.5 * j * t1 * t1;
        %t1 = sqrt(vel/j)
        vel = tc.cur_vel;
        % t = floor(sqrt(vel/tc.jerk) - 0.5);
        t = sqrt(vel/tc.jerk);
        % t = sqrt(vel/tc.jerk);
        % AT = AT + JT
        % t1 = floor(tc.maxaccel / tc.jerk - 0.5);   % max time for S4
        t1 = tc.maxaccel / tc.jerk;   % max time for S4
        % t1 = tc.maxaccel / tc.jerk;   % max time for S4
        if (t > t1) 
            % S4 . S5 . S6
            dist = tc.progress + t1 * vel;    % dist of (S4 + S6)
            % calc decel.dist for ACCEL_S5
            % t: time for S5
            % t = floor((vel - tc.maxaccel * t1) / tc.maxaccel - 0.5);
            t = (vel - tc.maxaccel * t1) / tc.maxaccel - 0.5;
            % t = (vel - tc.maxaccel * t1) / tc.maxaccel;
            v1 = vel - 0.5 * tc.jerk * t1 * t1;
            % PT = P0 + V0T + 1/2A0T^2 + 1/6JT^3
            dist = dist+ (v1 * t - 0.5 * tc.maxaccel * t * t);
        else 
            % S4 . S6
            dist = tc.progress + t * vel;    % dist of (S4 + S6)
        end
        
        % check if dist would be greater than tc_target at next cycle
        if (tc.tc_target < (dist - vel)) 
            tc.accel_state = tc.ACCEL_S4;
            % blending at largest velocity for G64 w/o P<tolerance>
            % if (!tc.tolerance) {
             %    tc.tolerance = tc.target - tc.progress; % tc.distance_to_go
            % } 
            break;
        end
        % check for changes of feed_override and request-velocity
        req_vel = tc.reqvel * tc.feed_override * tc.cycle_time;
        if (req_vel > tc.maxvel) 
            req_vel = tc.maxvel;
        end
    
        if ((tc.cur_vel + 1.5 * tc.jerk) < req_vel) 
            tc.accel_state = tc.ACCEL_S0;
            break;
        elseif ((tc.cur_vel - 1.5 * tc.jerk) > req_vel) 
            tc.accel_state = tc.ACCEL_S4;
            break;
        end
        tc.cur_vel = req_vel;
     

    case tc.ACCEL_S4
    % AT = AT + JT
    % VT = VT + AT + 1/2JT
    % PT = PT + VT + 1/2AT + 1/6JT
    tc.cur_accel = tc.cur_accel - tc.jerk;
    tc.cur_vel = tc.cur_vel + tc.cur_accel - 0.5 * tc.jerk;
    if (tc.cur_vel <= 0) 
        tc.cur_vel = 0;
        tc.accel_state = tc.ACCEL_S3;
        break;
    end
    tc.progress = tc.progress + tc.cur_vel + 0.5 * tc.cur_accel - 1.0/6.0 * tc.jerk;
    
    % (accel < 0) and (jerk < 0)    
    % check if we hit accel limit at next BP
    if ((tc.cur_accel - tc.jerk) <= -tc.maxaccel) 
        tc.cur_accel = -tc.maxaccel;
        tc.accel_state = tc.ACCEL_S5;
        break;
    end
    
    % should we stay in S4 and keep decel?
    % calculate dist for S6 . S4 . (maybe S5) . S6
    t = - tc.cur_accel / tc.jerk;
    % target dist after switching to S6 (jerk is positive for S6)
    dist = tc.progress + tc.cur_vel * t + 0.5 * tc.cur_accel * t * t + 1.0 / 6.0 * tc.jerk * t * t * t;
    % VT = V0 + A0T + 1/2JT2
    % obtain vel for S6 . S3
    vel = tc.cur_vel + tc.cur_accel * t + 0.5 * tc.jerk * t * t;
    
    if (vel > 0)  

        %  0.5 * vel = vel + 0 * t1 - 0.5 * j * t1 * t1;
        %  t1 = sqrt(vel/j)

        t = ceil(sqrt(vel/tc.jerk));
        % AT = AT + JT
        t1 = ceil(tc.maxaccel / tc.jerk);   % max time for S4
        if (t > t1) 
            % S4 . S5 . S6
            dist =dist+ t1 * vel;    % dist of (S4 + S6)
            % calc decel.dist for ACCEL_S5
            % t: time for S5
            t = (vel - tc.jerk * t1 * t1) / tc.maxaccel;
            v1 = vel - 0.5 * tc.jerk * t1 * t1;
            % PT = P0 + V0T + 1/2A0T^2 + 1/6JT^3
            dist =dist+  (v1 * t - 0.5 * tc.maxaccel * t * t);
        else 
            % S4 . S6
            dist =dist+  t * vel;    % dist of (S4 + S6)
        end
    end
    
    % check if dist would be greater than tc_target at next cycle
    if (tc.tc_target < (dist - (tc.cur_vel + 1.5 * tc.cur_accel - 2.1666667 * tc.jerk))) 
        tc.accel_state = tc.ACCEL_S4;
        %                DPS("should stay in S4 and keep decel\n");
        break;
    end
    
    % check if we will approaching requested velocity
    % vel should not be greater than "request velocity" after
    % starting acceleration to "accel == 0".
    %
    % AT = A0 + JT (let AT = 0 to calculate T)
    % VT = V0 + A0T + 1/2JT2
    t = - tc.cur_accel / tc.jerk;
    req_vel = tc.reqvel * tc.feed_override * tc.cycle_time;
    if (req_vel > tc.maxvel) 
        req_vel = tc.maxvel;
    end
    if ((tc.cur_vel + tc.cur_accel * t + 0.5 * tc.jerk * t * t) <= req_vel) 
        if(tc.progress/tc.target < 0.9)
            tc.accel_state = tc.ACCEL_S6;
        else
            tc.accel_state = tc.ACCEL_S7;
            tc.s6_v = tc.cur_vel;
            tc.s6_a = fabs(tc.cur_accel);
            tc.s6_p = tc.progress;
            tc.ts = floor((2*tc.s6_v)/tc.s6_a);
            tc.k = tc.s6_a*tc.pi/(4*tc.s6_v);
            tc.error_d = tc.target - tc.progress - tc.s6_v * tc.s6_v / tc.s6_a * (1-4/(tc.pi*tc.pi));
            tc.prev_s = 0;
            tc.prev_v = tc.s6_v;
            tc.c1 = -tc.s6_a/4;
            tc.c2 = tc.s6_v+((tc.error_d*tc.s6_a)/(2*tc.s6_v));
            tc.c3 = tc.s6_a/(8*tc.k*tc.k);
            tc.c4 = 2*tc.k;
            tc.c5 = -(tc.error_d*tc.s6_a)/(8*tc.k*tc.s6_v);
            tc.c6 = 4*tc.k;
            tc.ti = 1;
            break;
        end
    end

        
    case tc.ACCEL_S5
        % jerk is 0 at this state
        % accel < 0
        % VT = VT + AT + 1/2JT
        % PT = PT + VT + 1/2AT + 1/6JT
        tc.cur_vel = tc.cur_vel + tc.cur_accel;
        tc.progress = tc.progress + tc.cur_vel + 0.5 * tc.cur_accel;
        
        % should we stay in S5 and keep decel?
        % calculate dist for S6 . S4 . (maybe S5) . S6
        t = - tc.cur_accel / tc.jerk;
        % target dist after switching to S6 (jerk is positive for S6)
        dist = tc.progress + tc.cur_vel * t + 0.5 * tc.cur_accel * t * t+ 1.0 / 6.0 * tc.jerk * t * t * t;
        % VT = V0 + A0T + 1/2JT2
        % obtain vel for S6 . S3
        vel = tc.cur_vel + tc.cur_accel * t + 0.5 * tc.jerk * t * t;
        
        if (vel > 0)  
            %/* S6 . S3 . S4 . S5(maybe) . S6 */
            %/* 
            % 0.5 * vel = vel + 0 * t1 - 0.5 * j * t1 * t1;
            %t1 = sqrt(vel/j)
            %*/
            t = ceil(sqrt(vel/tc.jerk));
            % AT = AT + JT
            t1 = ceil(tc.maxaccel / tc.jerk);   % max time for S4
            if (t > t1) 
                % S4 . S5 . S6
                dist =dist+ t1 * vel;    % dist of (S4 + S6)
                % calc decel.dist for ACCEL_S5
                % t: time for S5
                t = (vel - tc.jerk * t1 * t1) / tc.maxaccel;
                v1 = vel - 0.5 * tc.jerk * t1 * t1;
                % PT = P0 + V0T + 1/2A0T^2 + 1/6JT^3
                dist =dist+ (v1 * t - 0.5 * tc.maxaccel * t * t);
            else 
                % S4 . S6
                dist =dist+ t * vel;    % dist of (S4 + S6)
            end
        end
        
        % check if dist would be greater than tc_target at next cycle
        if (tc.tc_target < (dist - (tc.cur_vel + 1.5 * tc.cur_accel))) 
            tc.accel_state = tc.ACCEL_S5;
            %                DPS("should stay in S5 and keep decel\n");
            break;
        end
        
        % check if we will approaching requested velocity
        % vel should not be greater than "request velocity" after
        % starting acceleration to "accel == 0".
        %
        % AT = A0 + JT (let AT = 0 to calculate T)
        % VT = V0 + A0T + 1/2JT2
        % t: cycles for accel to decel to 0
        t = - tc.cur_accel / tc.jerk;
        req_vel = tc.reqvel * tc.feed_override * tc.cycle_time;
        if (req_vel > tc.maxvel)    
            req_vel = tc.maxvel;
        end
        if ((tc.cur_vel + tc.cur_accel * t + 0.5 * tc.jerk * t * t) <= req_vel) 
            if(tc.progress/tc.target < 0.9)
                tc.accel_state = tc.ACCEL_S6;
            else
                tc.accel_state = tc.ACCEL_S7;
                tc.s6_v = tc.cur_vel;
                tc.s6_a = abs(tc.cur_accel);
                tc.s6_p = tc.progress;
                tc.ts = floor((2*tc.s6_v)/tc.s6_a);
                tc.k = tc.s6_a*tc.pi/(4*tc.s6_v);
                tc.error_d = tc.target - tc.progress - tc.s6_v * tc.s6_v / tc.s6_a * (1-4/(tc.pi*tc.pi));
                tc.prev_s = 0;
                tc.prev_v = tc.s6_v;
                tc.c1 = -tc.s6_a/4;
                tc.c2 = tc.s6_v+((tc.error_d*tc.s6_a)/(2*tc.s6_v));
                tc.c3 = tc.s6_a/(8*tc.k*tc.k);
                tc.c4 = 2*tc.k;
                tc.c5 = -(tc.error_d*tc.s6_a)/(8*tc.k*tc.s6_v);
                tc.c6 = 4*tc.k;
                tc.ti = 1;
                break;
            end
        end
        
    case tc.ACCEL_S6
        % AT = AT + JT
        % VT = VT + AT + 1/2JT
        % PT = PT + VT + 1/2AT + 1/6JT
        req_vel = tc.reqvel * tc.feed_override * tc.cycle_time;
        if (req_vel > tc.maxvel) 
            req_vel = tc.maxvel;
        end
        
        tc.cur_accel = tc.cur_accel + tc.jerk;
        tc.cur_vel = tc.cur_vel + tc.cur_accel + 0.5 * tc.jerk;
        
        if (tc.cur_vel <= req_vel) 
            tc.accel_state = tc.ACCEL_S3;
            if ((req_vel - tc.cur_vel) < 1.5*tc.jerk) 
                % align to req_vel only when not changing feed_override
                tc.cur_vel = req_vel;
            end
        end
        dist = tc.cur_vel + 0.5 * tc.cur_accel + 1.0/6.0 * tc.jerk;
        tc.progress = tc.progress + dist;
        
        if (tc.cur_accel >= 0) 
            tc.accel_state = tc.ACCEL_S3;
        end
    case tc.ACCEL_S7
        % decel to target position based on Jofey's algorithm
        
        if(tc.ti <= tc.ts)
            dist = tc.c1*tc.ti*tc.ti + tc.c2*tc.ti + tc.c3*cos(tc.c4*tc.ti) + tc.c5*cos(tc.c6*tc.ti-0.5*tc.pi) - tc.c3;
            tc.cur_vel = dist - tc.prev_s;
            tc.cur_accel = tc.cur_vel - tc.prev_v;
            tc.prev_s = dist;
            tc.prev_v = tc.cur_vel;
            tc.progress = tc.s6_p + dist;
            tc.ti = tc.ti + 1;
        else 
            tc.cur_vel = 0;
            tc.cur_accel = 0;
            tc.progress = tc.target;
            tc.accel_state = tc.ACCEL_S3;
            tc.isend = 1;
        end

    end %switch

   flag = 0 ;
   end

   
end
